"""Functions intended to be passed to be dask.

These will typically be functions that operate on dask chunks. For
instance, as would be passed to `dask.array.map_blocks`.

"""

import numpy as np


def cfdm_to_memory(a):
    """Return an in-memory version of *a*.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        a: array_like
            The array.

    :Returns:

            If *a* has an `__in_memory__` attribute that is False then
            ``np.asanyarray(a)`` will be returned. Otherwise *a* will
            be returned unchanged.

    """
    if not getattr(a, "__in_memory__", True):
        return np.asanyarray(a)

    return a


def cfdm_filled(a, fill_value=None):
    """Replace masked elements with a fill value.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        a: array_like
            The array.

        fill_value: scalar
            The fill value.

    :Returns:

        `numpy.ndarray`
            The filled array.

    **Examples**

    >>> a = np.array([[1, 2, 3]])
    >>> print(data.dask_utils.filled(a, -999))
    [[1 2 3]]
    >>> a = np.ma.array([[1, 2, 3]], mask=[[True, False, False]])
    >>> print(data.dask_utils.filled(a, -999))
    [[-999    2    3]]

    """
    a = cfdm_to_memory(a)
    return np.ma.filled(a, fill_value=fill_value)


def cfdm_harden_mask(a):
    """Harden the mask of a masked `numpy` array.

    Has no effect if the array is not a masked array.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        a: `numpy.ndarray`
            The array to give a hardened mask.

    :Returns:

        `numpy.ndarray`
            The array with hardened mask.

    """
    a = cfdm_to_memory(a)
    if np.ma.isMA(a):
        try:
            a.harden_mask()
        except AttributeError:
            # Trap cases when the input array is not a numpy array
            # (e.g. it might be numpy.ma.masked).
            pass

    return a


def cfdm_soften_mask(a):
    """Soften the mask of a masked `numpy` array.

    Has no effect if the array is not a masked array.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        a: `numpy.ndarray`
            The array to give a softened mask.

    :Returns:

        `numpy.ndarray`
            The array with softened mask.

    """
    a = cfdm_to_memory(a)
    if np.ma.isMA(a):
        try:
            a.soften_mask()
        except AttributeError:
            # Trap cases when the input array is not a numpy array
            # (e.g. it might be numpy.ma.masked).
            pass

    return a


def cfdm_where(array, condition, x, y, hardmask):
    """Set elements of *array* from *x* or *y* depending on *condition*.

    The input *array* is not changed in-place.

    **Missing data**

    Array elements may be set to missing values if either *x* or *y*
    are the `cfdm.masked` constant, or by assignment from any missing
    data elements in *x* or *y*.

    If the *array* mask is hard then missing data values in the array
    will not be overwritten, regardless of the content of *x* and *y*.

    If the *condition* contains missing data then the corresponding
    elements in the array will not be assigned to, regardless of the
    contents of *x* and *y*.

    .. note:: This function sets the mask hardness of the output
              array appropriately.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        array: numpy.ndarray
            The array to be assigned to.

        condition: numpy.ndarray
            Where False or masked, assign from *y*, otherwise assign
            from *x*.

        x: numpy.ndarray or `None`
            *x* and *y* must not both be `None`.

        y: numpy.ndarray or `None`
            *x* and *y* must not both be `None`.

        hardmask: `bool`
           Set the mask hardness for a returned masked `numpy` array.

           If False then a returned masked `numpy` array will have a
           soft mask.

           If True then mask of the input `numpy` *array* (if there is
           one) will be applied to the returned `numpy` array, in
           addition to any masked elements arising from assignments
           from *x* or *y*; and the if the returned `numpy` array is
           masked then it will have a hard mask.

    :Returns:

        `numpy.ndarray`
            A copy of the input *array* with elements from *y* where
            *condition* is False or masked, and elements from *x*
            elsewhere.

    """
    array = cfdm_to_memory(array)
    condition = cfdm_to_memory(condition)
    if x is not None:
        x = cfdm_to_memory(x)

    if y is not None:
        y = cfdm_to_memory(y)

    mask = None

    if np.ma.isMA(array):
        # Do a masked where
        where = np.ma.where
        if hardmask:
            mask = array.mask
    elif np.ma.isMA(x) or np.ma.isMA(y):
        # Do a masked where
        where = np.ma.where
    else:
        # Do a non-masked where
        where = np.where
        hardmask = False

    condition_is_masked = np.ma.isMA(condition)
    if condition_is_masked:
        condition = condition.astype(bool)

    if x is not None:
        # Assign values from x
        if condition_is_masked:
            # Replace masked elements of condition with False, so that
            # masked locations are assigned from array
            c = condition.filled(False)
        else:
            c = condition

        array = where(c, x, array)

    if y is not None:
        # Assign values from y
        if condition_is_masked:
            # Replace masked elements of condition with True, so that
            # masked locations are assigned from array
            c = condition.filled(True)
        else:
            c = condition

        array = where(c, array, y)

    if hardmask:
        if mask is not None and mask.any():
            # Apply the mask from the input array to the result
            array.mask |= mask

        array.harden_mask()

    return array



import logging

from ..decorators import (
    _display_or_return,
    _inplace_enabled,
    _inplace_enabled_define_and_cleanup,
    _manage_log_level_via_verbosity,
)
from ..functions import parse_indices
from . import PropertiesData

logger = logging.getLogger(__name__)


class PropertiesDataBounds(PropertiesData):
    """Mixin for a data array with descriptive properties and bounds.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        properties=None,
        data=None,
        bounds=None,
        geometry=None,
        interior_ring=None,
        node_count=None,
        part_node_count=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            properties: `dict`, optional
                Set descriptive properties. The dictionary keys are
                property names, with corresponding values.

                Properties may also be set after initialisation with
                the `properties` and `set_property` methods.

                *Parameter example:*
                  ``properties={'standard_name': 'longitude'}``

            {{init data: data_like, optional}}

            bounds: `Bounds`, optional
                Set the bounds array.

                The bounds array may also be set after initialisation
                with the `set_bounds` method.

            geometry: `str`, optional
                Set the geometry type.

                The geometry type may also be set after initialisation
                with the `set_geometry` method.

                *Parameter example:*
                  ``geometry='polygon'``

            interior_ring: `InteriorRing`, optional
                Set the interior ring variable.

                The interior ring variable may also be set after
                initialisation with the `set_interior_ring` method.

            node_count: `NodeCount`, optional
                Set the node count variable for geometry
                bounds.

                The node count variable may also be set after
                initialisation with the `set_node_count` method.

            part_node_count: `PartNodeCount`, optional
                Set the part node count variable for geometry
                bounds.

                The part node count variable may also be set after
                initialisation with the `set_node_count` method.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        # Initialise properties, data, geometry and interior ring
        super().__init__(
            properties=properties,
            data=data,
            bounds=bounds,
            source=source,
            geometry=geometry,
            interior_ring=interior_ring,
            copy=copy,
            _use_data=_use_data,
        )

        # Get node count and part node count variables from source
        if source is not None:
            try:
                node_count = source.get_node_count(None)
            except AttributeError:
                node_count = None

            try:
                part_node_count = source.get_part_node_count(None)
            except AttributeError:
                part_node_count = None

        # Initialise node count
        if node_count is not None:
            self.set_node_count(node_count, copy=copy)

        # Initialise part node count
        if part_node_count is not None:
            self.set_part_node_count(part_node_count, copy=copy)

    def __getitem__(self, indices):
        """Return a subspace of the construct defined by indices.

        f.__getitem__(indices) <==> f[indices]

        The new subspace contains the same properties and similar
        components to the original construct, but the latter are also
        subspaced over their corresponding axes.

        Indexing follows rules that are very similar to the numpy
        indexing rules, the only differences being:

        * An integer index i takes the i-th element but does not
          reduce the rank by one.

        * When two or more dimensions' indices are sequences of integers
          then these indices work independently along each dimension
          (similar to the way vector subscripts work in Fortran). This is
          the same behaviour as indexing on a Variable object of the
          netCDF4 package.

        :Returns:

            `{{class}}`
                The subspace of the construct.

        **Examples**

        >>> f.shape
        (1, 10, 9)
        >>> f[:, :, 1].shape
        (1, 10, 1)
        >>> f[:, 0].shape
        (1, 1, 9)
        >>> f[..., 6:3:-1, 3:6].shape
        (1, 3, 3)
        >>> f[0, [2, 9], [4, 8]].shape
        (1, 2, 2)
        >>> f[0, :, -2].shape
        (1, 10, 1)

        """
        if indices is Ellipsis:
            return self.copy()

        if not isinstance(indices, tuple):
            indices = (indices,)

        new = super().__getitem__(indices)

        # Subspace the interior ring array, if there is one.
        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            new.set_interior_ring(interior_ring[indices], copy=False)

        # Subspace the bounds, if there are any.
        self_bounds = self.get_bounds(None)
        if self_bounds is not None:
            data = self_bounds.get_data(None, _units=False, _fill_value=False)
            if data is not None:
                # There is a bounds array
                bounds_indices = list(parse_indices(data.shape, indices))

                if data.ndim <= 2:
                    index = bounds_indices[0]
                    if isinstance(index, slice):
                        if index.step and index.step < 0:
                            # This scalar or 1-d variable has been
                            # reversed so reverse its bounds (as per
                            # 7.1 of the conventions)
                            bounds_indices[-1] = slice(None, None, -1)
                    elif data.size > 1 and int(index[-1]) < int(index[0]):
                        # This 1-d variable has been reversed so
                        # reverse its bounds (as per 7.1 of the
                        # conventions)
                        bounds_indices[-1] = slice(None, None, -1)

                new.set_bounds(self_bounds[tuple(bounds_indices)], copy=False)

        # Return the new bounded variable
        return new

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        shape = None
        data = self.get_data(None, _units=False, _fill_value=False)
        bounds = self.get_bounds(None)
        if data is not None:
            shape = data.shape

        if shape is not None:
            dims = ", ".join([str(x) for x in shape])
            dims = f"({dims})"
        else:
            dims = ""

        # ------------------------------------------------------------
        # Units and calendar
        # ------------------------------------------------------------

        units = self.get_property("units", None)
        if units is None and bounds is not None:
            units = bounds.get_property("units", None)

        calendar = self.get_property("calendar", None)
        if calendar is None and bounds is not None:
            calendar = bounds.get_property("calendar", None)

        if units is None:
            isreftime = calendar is not None
            units = ""
        else:
            isreftime = "since" in units

        if isreftime:
            if calendar is None:
                calendar = ""

            units += " " + calendar

        return f"{self.identity('')}{dims} {units}"

    def _original_filenames(self, define=None, update=None, clear=False):
        """The names of files containing the original data and metadata.

        {{original filenames}}

        .. versionadded:: (cfdm) 1.10.0.1

        :Parameters:

            {{define: (sequence of) `str`, optional}}

            {{update: (sequence of) `str`, optional}}

            {{clear: `bool` optional}}

        :Returns:

            `set`
                {{Returns original filenames}}

                If the *define* or *update* parameter is set then
                `None` is returned.

        """
        out = super()._original_filenames(
            define=define, update=update, clear=clear
        )
        if out is None:
            return

        bounds = self.get_bounds(None)
        if bounds is not None:
            out.update(bounds._original_filenames(clear=clear))

        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            out.update(interior_ring._original_filenames(clear=clear))

        return out

    @_inplace_enabled(default=False)
    def apply_masking(self, bounds=True, inplace=False):
        """Apply masking as defined by the CF conventions.

        Masking is applied according to any of the following criteria
        that are applicable:

        * where data elements are equal to the value of the
          ``missing_value`` property;

        * where data elements are equal to the value of the
          ``_FillValue`` property;

        * where data elements are strictly less than the value of the
          ``valid_min`` property;

        * where data elements are strictly greater than the value of
          the ``valid_max`` property;

        * where data elements are within the inclusive range specified
          by the two values of ``valid_range`` property.

        If any of the above properties have not been set the no
        masking is applied for that method.

        The cell bounds, if any, are also masked according to the same
        criteria as the parent construct. If, however, any of the
        relevant properties are explicitly set on the bounds instance
        then their values will be used in preference to those of the
        parent construct.

        Elements that are already masked remain so.

        .. note:: If using the `apply_masking` method on a construct
                  that has been read from a dataset with the
                  ``mask=False`` parameter to the `read` function,
                  then the mask defined in the dataset can only be
                  recreated if the ``missing_value``, ``_FillValue``,
                  ``valid_min``, ``valid_max``, and ``valid_range``
                  properties have not been updated.

        .. versionadded:: (cfdm) 1.8.2

        .. seealso:: `Data.apply_masking`, `read`, `write`

        :Parameters:

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                A new instance with masked values, or `None` if the
                operation was in-place.

        **Examples**

        >>> print(c.data.array)
        [9.96920997e+36, 9.96920997e+36, 9.96920997e+36, 9.96920997e+36,
         9.96920997e+36, 9.96920997e+36, 9.96920997e+36, 9.96920997e+36],
         [0.023 0.036 0.045 0.062 0.046 0.073 0.006 0.066]
         [0.11  0.131 0.124 0.146 0.087 0.103 0.057 0.011]
         [0.029 0.059 0.039 0.07  0.058 0.072 0.009 0.017]
        [9.96920997e+36, 9.96920997e+36, 9.96920997e+36, 9.96920997e+36,
         9.96920997e+36, 9.96920997e+36, 9.96920997e+36, 9.96920997e+36]])
        >>> masked_c = c.apply_masking()
        >>> print(masked_c.data.array)
        [[   --    --    --    --    --    --    --    --]
         [0.023 0.036 0.045 0.062 0.046 0.073 0.006 0.066]
         [0.11  0.131 0.124 0.146 0.087 0.103 0.057 0.011]
         [0.029 0.059 0.039 0.07  0.058 0.072 0.009 0.017]
         [   --    --    --    --    --    --    --    --]]

        """
        c = _inplace_enabled_define_and_cleanup(self)
        super(PropertiesDataBounds, c).apply_masking(inplace=True)

        data = c.get_bounds_data(None, _units=False, _fill_value=False)
        if data is not None:
            b = c.get_bounds()

            fill_values = []
            for prop in ("_FillValue", "missing_value"):
                x = b.get_property(prop, c.get_property(prop, None))
                if x is not None:
                    fill_values.append(x)

            kwargs = {"inplace": True, "fill_values": fill_values}

            for prop in ("valid_min", "valid_max", "valid_range"):
                kwargs[prop] = b.get_property(prop, c.get_property(prop, None))

            data.apply_masking(**kwargs)

        return c

    @classmethod
    def concatenate(
        cls,
        variables,
        axis=0,
        cull_graph=False,
        relaxed_units=False,
        copy=True,
    ):
        """Join a together sequence of `{{class}}`.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `Data.concatenate`, `Data.cull_graph`

        :Parameters:

            variables: sequence of constructs

            axis: `int`, optional
                Select the axis to along which to concatenate, defined
                by its position in the data array. By default
                concatenation is along the axis in position 0.

            {{cull_graph: `bool`, optional}}

            {{relaxed_units: `bool`, optional}}

            {{concatenate copy: `bool`, optional}}

        :Returns:

            `{{class}}`
                The concatenated construct.

        """
        variable0 = variables[0]
        if copy:
            variable0 = variable0.copy()

        if len(variables) == 1:
            return variable0

        out = super().concatenate(
            variables,
            axis=axis,
            cull_graph=cull_graph,
            relaxed_units=relaxed_units,
            copy=copy,
        )

        bounds = variable0.get_bounds(None)
        if bounds is not None:
            bounds = bounds.concatenate(
                [v.get_bounds() for v in variables],
                axis=axis,
                cull_graph=cull_graph,
                relaxed_units=relaxed_units,
                copy=copy,
            )
            out.set_bounds(bounds, copy=False)

        interior_ring = variable0.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring = interior_ring.concatenate(
                [v.get_interior_ring() for v in variables],
                axis=axis,
                cull_graph=cull_graph,
                relaxed_units=relaxed_units,
                copy=copy,
            )
            out.set_interior_ring(interior_ring, copy=False)

        return out

    def creation_commands(
        self,
        representative_data=False,
        namespace=None,
        indent=0,
        string=True,
        name="c",
        data_name="data",
        bounds_name="b",
        interior_ring_name="i",
        header=True,
        _coordinate=False,
    ):
        """Return the commands that would create the construct.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `{{package}}.Data.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{representative_data: `bool`, optional}}

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{data_name: `str`, optional}}

            bounds_name: `str`, optional
                The name of the construct's `Bounds` instance created
                by the returned commands.

                *Parameter example:*
                  ``name='bounds1'``

            interior_ring_name: `str`, optional
                The name of the construct's `InteriorRing` instance
                created by the returned commands.

                *Parameter example:*
                  ``name='ir1'``

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> x = {{package}}.{{class}}(
        ...     properties={'units': 'degrees_east',
        ...                 'standard_name': 'longitude'}
        ... )
        >>> x.set_data([22.5, 67.5, 112.5])
        >>> b = {{package}}.Bounds()
        >>> b.set_data([[0.0, 45.0], [45.0, 90.0], [90.0, 135.0]])
        >>> x.set_bounds(b)
        >>> print(x.creation_commands(header=False))
        c = {{package}}.{{class}}()
        c.set_properties({'units': 'degrees_east', 'standard_name': 'longitude'})
        data = {{package}}.Data([22.5, 67.5, 112.5], units='degrees_east', dtype='f8')
        c.set_data(data)
        b = {{package}}.Bounds()
        data = {{package}}.Data([[0.0, 45.0], [45.0, 90.0], [90.0, 135.0]], units='degrees_east', dtype='f8')
        b.set_data(data)
        c.set_bounds(b)

        """
        if name in (data_name, bounds_name, interior_ring_name):
            raise ValueError(
                "The 'name' parameter can not have the same value as "
                "any of the 'data_name', 'bounds_name', or "
                f"'interior_ring_name' parameters: {name!r}"
            )

        if data_name in (name, bounds_name, interior_ring_name):
            raise ValueError(
                "The 'data_name' parameter can not have "
                "the same value as any of the 'name', 'bounds_name', "
                f"or 'interior_ring_name' parameters: {data_name!r}"
            )

        namespace0 = namespace
        if namespace is None:
            namespace = self._package() + "."
        elif namespace and not namespace.endswith("."):
            namespace += "."

        out = super().creation_commands(
            representative_data=representative_data,
            indent=0,
            namespace=namespace,
            string=False,
            name=name,
            data_name=data_name,
            header=header,
        )

        # Geometry type
        geometry = self.get_geometry(None)
        if geometry is not None:
            out.append(f"{name}.set_geometry({geometry!r})")

        # Climatology
        if _coordinate and self.get_climatology(False):
            out.append("{}.set_climatology(True)".format(name))

        bounds = self.get_bounds(None)
        if bounds is not None:
            out.extend(
                bounds.creation_commands(
                    representative_data=representative_data,
                    indent=0,
                    namespace=namespace0,
                    string=False,
                    name=bounds_name,
                    data_name=data_name,
                    header=False,
                )
            )
            out.append(f"{name}.set_bounds({bounds_name})")

        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            out.extend(
                interior_ring.creation_commands(
                    representative_data=representative_data,
                    indent=0,
                    namespace=namespace0,
                    string=False,
                    name=interior_ring_name,
                    data_name=data_name,
                    header=False,
                )
            )
            out.append(f"{name}.set_interior_ring({interior_ring_name})")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    def del_node_count(self, default=ValueError()):
        """Remove the node count variable for geometry bounds.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `get_node_count`, `has_node_count`,
                     `set_node_count`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                node count variable has not been set.

                {{default Exception}}

        :Returns:

            `NodeCount`
                The removed node count variable.

        **Examples**

        >>> n = {{package}}.NodeCount(properties={'long_name': 'node counts'})
        >>> c.set_node_count(n)
        >>> c.has_node_count()
        True
        >>> c.get_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.del_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.has_node_count()
        False

        """
        try:
            return self._del_component("node_count")
        except ValueError:
            if default is None:
                return default

            return self._default(
                default,
                f"{self.__class__.__name__} has no node count variable",
            )

    def del_part_node_count(self, default=ValueError()):
        """Remove the part node count variable for geometry bounds.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `get_part_node_count`, `has_part_node_count`,
                     `set_part_node_count`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the part
                node count variable has not been set.

                {{default Exception}}

        :Returns:

            `PartNodeCount`
                The removed part node count variable.

        **Examples**

        >>> p = {{package}}.PartNodeCount(properties={'long_name': 'part node counts'})
        >>> c.set_part_node_count(p)
        >>> c.has_part_node_count()
        True
        >>> c.get_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.del_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.has_part_node_count()
        False

        """
        try:
            return self._del_component("part_node_count")
        except ValueError:
            if default is None:
                return default

            return self._default(
                default,
                f"{self.__class__.__name__} has no part node count variable",
            )

    @_display_or_return
    def dump(
        self,
        display=True,
        _key=None,
        _omit_properties=None,
        _prefix="",
        _title=None,
        _create_title=True,
        _level=0,
        _axes=None,
        _axis_names=None,
    ):
        """A full description.

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default

        :Returns:

            {{returns dump}}

        """
        # ------------------------------------------------------------
        # Properties and Data
        # ------------------------------------------------------------
        string = super().dump(
            display=False,
            _key=_key,
            _omit_properties=_omit_properties,
            _prefix=_prefix,
            _title=_title,
            _create_title=_create_title,
            _level=_level,
            _axes=_axes,
            _axis_names=_axis_names,
        )

        string = [string]

        # ------------------------------------------------------------
        # Geometry type
        # ------------------------------------------------------------
        geometry = self.get_geometry(None)
        if geometry is not None:
            indent1 = "    " * (_level + 1)
            string.append(f"{indent1}{_prefix}Geometry: {geometry}")

        # ------------------------------------------------------------
        # Bounds
        # ------------------------------------------------------------
        bounds = self.get_bounds(None)
        if bounds is not None:
            string.append(
                bounds.dump(
                    display=False,
                    _key=_key,
                    _prefix=_prefix + "Bounds:",
                    _create_title=False,
                    _level=_level,
                    _axes=_axes,
                    _axis_names=_axis_names,
                )
            )

        # -------------------------------------------------------------
        # Interior ring
        # ------------------------------------------------------------
        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            string.append(
                interior_ring.dump(
                    display=False,
                    _key=_key,
                    _prefix=_prefix + "Interior Ring:",
                    _create_title=False,
                    _level=_level,
                    _axes=_axes,
                    _axis_names=_axis_names,
                )
            )

        return "\n".join(string)

    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_data_type=False,
        ignore_fill_value=False,
        ignore_properties=None,
        ignore_compression=True,
        ignore_type=False,
    ):
        """Whether two instances are the same.

        Equality is strict by default. This means that:

        * the same descriptive properties must be present, with the
          same values and data types, and vector-valued properties
          must also have same the size and be element-wise equal (see
          the *ignore_properties* and *ignore_data_type* parameters),
          and

        ..

        * if there are data arrays then they must have same shape and
          data type, the same missing data mask, and be element-wise
          equal (see the *ignore_data_type* parameter).

        ..

        * if there are bounds then their descriptive properties (if
          any) must be the same and their data arrays must have same
          shape and data type, the same missing data mask, and be
          element-wise equal (see the *ignore_properties* and
          *ignore_data_type* parameters).

        {{equals tolerance}}

        {{equals compression}}

        {{equals netCDF}}

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:


            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            {{ignore_fill_value: `bool`, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

            {{ignore_properties: (sequence of) `str`, optional}}

            {{ignore_data_type: `bool`, optional}}

            {{ignore_compression: `bool`, optional}}

            {{ignore_type: `bool`, optional}}

        :Returns:

            `bool`
                Whether the two instances are equal.

        **Examples**

        >>> p.equals(p)
        True
        >>> p.equals(p.copy())
        True
        >>> p.equals('not a colection of properties')
        False

        """
        # ------------------------------------------------------------
        # Check the properties and data
        # ------------------------------------------------------------
        if not super().equals(
            other,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_data_type=ignore_data_type,
            ignore_fill_value=ignore_fill_value,
            ignore_properties=ignore_properties,
            ignore_type=ignore_type,
            ignore_compression=ignore_compression,
        ):
            return False

        # ------------------------------------------------------------
        # Check the geometry type
        # ------------------------------------------------------------
        if self.get_geometry(None) != other.get_geometry(None):
            logger.info(
                f"{self.__class__.__name__}: Different geometry types: "
                f"{self.get_geometry(None)}, {other.get_geometry(None)}"
            )
            return False

        # ------------------------------------------------------------
        # Check the bounds
        # ------------------------------------------------------------
        self_has_bounds = self.has_bounds()
        if self_has_bounds != other.has_bounds():
            logger.info(f"{self.__class__.__name__}: Different bounds")
            return False

        if self_has_bounds:
            if not self._equals(
                self.get_bounds(),
                other.get_bounds(),
                rtol=rtol,
                atol=atol,
                verbose=verbose,
                ignore_data_type=ignore_data_type,
                ignore_type=ignore_type,
                ignore_fill_value=ignore_fill_value,
                ignore_compression=ignore_compression,
            ):
                logger.info(
                    f"{self.__class__.__name__}: Different bounds"
                )  # pragma: no cover

                return False

        # ------------------------------------------------------------
        # Check the interior ring
        # ------------------------------------------------------------
        self_has_interior_ring = self.has_interior_ring()
        if self_has_interior_ring != other.has_interior_ring():
            logger.info(
                f"{self.__class__.__name__}: Different interior ring"
            )  # pragma: no cover

            return False

        if self_has_interior_ring:
            if not self._equals(
                self.get_interior_ring(),
                other.get_interior_ring(),
                rtol=rtol,
                atol=atol,
                verbose=verbose,
                ignore_data_type=ignore_data_type,
                ignore_type=ignore_type,
                ignore_fill_value=ignore_fill_value,
                ignore_compression=ignore_compression,
            ):
                logger.info(
                    f"{self.__class__.__name__}: Different interior ring"
                )  # pragma: no cover

                return False

        return True

    def get_node_count(self, default=ValueError()):
        """Return the node count variable for geometry bounds.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `del_node_count`, `has_node_count`, `set_node_count`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if a node
                count variable has not been set.

                {{default Exception}}

        :Returns:

            `NodeCount`
                The node count variable.

        **Examples**

        >>> n = {{package}}.NodeCount(properties={'long_name': 'node counts'})
        >>> c.set_node_count(n)
        >>> c.has_node_count()
        True
        >>> c.get_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.del_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.has_node_count()
        False

        """
        out = self._get_component("node_count", None)
        if out is None:
            if default is None:
                return default

            return self._default(
                default,
                f"{self.__class__.__name__} has no node count variable",
            )

        return out

    def file_directories(self):
        """The directories of files containing parts of the data.

        Returns the locations of any files referenced by the data.

        .. seealso:: `get_filenames`, `replace_directory`

        :Returns:

            `set`
                The unique set of file directories as absolute paths.

        **Examples**

        >>> d.file_directories()
        {'https:///data/1', 'file:///data2'}

        """
        out = super().file_directories()

        bounds = self.get_bounds(None)
        if bounds is not None:
            out.update(bounds.file_directories())

        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            out.update(interior_ring.file_directories())

        return out

    def get_part_node_count(self, default=ValueError()):
        """Return the part node count variable for geometry bounds.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `del_part_node_count`, `get_node_count`,
                     `has_part_node_count`, `set_part_node_count`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the part
                node count variable has not been set.

                {{default Exception}}

        :Returns:

            `PartNodeCount`
                The part node count variable.

        **Examples**

        >>> p = {{package}}.PartNodeCount(properties={'long_name': 'part node counts'})
        >>> c.set_part_node_count(p)
        >>> c.has_part_node_count()
        True
        >>> c.get_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.del_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.has_part_node_count()
        False

        """
        out = self._get_component("part_node_count", None)
        if out is None:
            if default is None:
                return default

            return self._default(
                default,
                f"{self.__class__.__name__} has no part node count variable",
            )

        return out

    def has_node_count(self):
        """Whether geometry bounds have a node count variable.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `del_node_count`, `get_node_count`, `set_node_count`

        :Returns:

            `bool`
                True if there is a node count variable, otherwise False.

        **Examples**


        >>> n = {{package}}.NodeCount(properties={'long_name': 'node counts'})
        >>> c.set_node_count(n)
        >>> c.has_node_count()
        True
        >>> c.get_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.del_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.has_node_count()
        False

        """
        return self._has_component("node_count")

    def has_part_node_count(self):
        """Whether geometry bounds have a part node count variable.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `del_part_node_count`, `get_part_node_count`,
                     `set_part_node_count`

        :Returns:

            `bool`
                True if there is a part node count variable, otherwise
                False.

        **Examples**

        >>> p = {{package}}.PartNodeCount(properties={'long_name': 'part node counts'})
        >>> c.set_part_node_count(p)
        >>> c.has_part_node_count()
        True
        >>> c.get_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.del_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.has_part_node_count()
        False

        """
        return self._has_component("part_node_count")

    def identities(self, generator=False, **kwargs):
        """Return all possible identities.

        The identities comprise:

        * The ``standard_name`` property.
        * All properties, preceded by the property name and a colon,
          e.g. ``'long_name:Air temperature'``.
        * The netCDF variable name, preceded by ``'ncvar%'``.
        * The identities of the bounds, if any.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list.

                .. versionadded:: (cfdm) 1.8.9.0

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `list` or generator
                The identities.

        **Examples**

        >>> f.properties()
        {'foo': 'bar',
         'long_name': 'Air Temperature',
         'standard_name': 'air_temperature'}
        >>> f.nc_get_variable()
        'tas'
        >>> f.identities()
        ['air_temperature',
         'long_name=Air Temperature',
         'foo=bar',
         'standard_name=air_temperature',
         'ncvar%tas']

        >>> f.properties()
        {}
        >>> f.bounds.properties()
        {'axis': 'Z',
         'units': 'm'}
        >>> f.identities()
        ['axis=Z', 'units=m', 'ncvar%z']
        >>> for i in f.identities(generator=True):
        ...     print(i)
        ...
        axis=Z
        units=m
        ncvar%z

        """
        bounds = self.get_bounds(None)
        if bounds is not None:
            post = (bounds.identities(generator=True),)
            post0 = kwargs.pop("post", None)
            if post0:
                post = tuple(post0) + post

            kwargs["post"] = post

        return super().identities(generator=generator, **kwargs)

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        1. The ``standard_name`` property.
        2. The ``cf_role`` property, preceded by ``'cf_role='``.
        3. The ``axis`` property, preceded by ``'axis='``.
        4. The ``long_name`` property, preceded by ``'long_name='``.
        5. The netCDF variable name, preceded by ``'ncvar%'``.
        6. The identity of the bounds, if any.
        7. The value of the *default* parameter.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> f.properties()
        {'foo': 'bar',
         'long_name': 'Air Temperature',
         'standard_name': 'air_temperature'}
        >>> f.nc_get_variable()
        'tas'
        >>> f.identity()
        'air_temperature'
        >>> f.del_property('standard_name')
        'air_temperature'
        >>> f.identity(default='no identity')
        'air_temperature'
        >>> f.identity()
        'long_name=Air Temperature'
        >>> f.del_property('long_name')
        >>> f.identity()
        'ncvar%tas'
        >>> f.nc_del_variable()
        'tas'
        >>> f.identity()
        'ncvar%tas'
        >>> f.identity()
        ''
        >>> f.identity(default='no identity')
        'no identity'

        >>> f.properties()
        {}
        >>> f.bounds.properties()
        {'axis': 'Z',
         'units': 'm'}
        >>> f.identity()
        'axis=Z'

        """
        identity = super().identity(default=None)
        if identity is not None:
            return identity

        bounds = self.get_bounds(None)
        if bounds is not None:
            return bounds.identity(default=default)

        return default

    def get_bounds(self, default=ValueError()):
        """Return the bounds.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `bounds`, `get_data`, `del_bounds`, `has_bounds`,
                     `set_bounds`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if bounds have
                not been set.

                {{default Exception}}

        :Returns:

            `Bounds`
                The bounds.

        **Examples**

        >>> b = {{package}}.Bounds(data={{package}}.Data(range(10).reshape(5, 2)))
        >>> c.set_bounds(b)
        >>> c.has_bounds()
        True
        >>> c.get_bounds()
        <{{repr}}Bounds: (5, 2) >
        >>> b = c.del_bounds()
        >>> b
        <{{repr}}Bounds: (5, 2) >
        >>> c.has_bounds()
        False
        >>> print(c.get_bounds(None))
        None
        >>> print(c.del_bounds(None))
        None

        """
        bounds = super().get_bounds(default=None)

        if bounds is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no bounds"
            )

        bounds._set_component(
            "inherited_properties", self.properties(), copy=False
        )

        return bounds

    def get_bounds_data(
        self, default=ValueError(), _units=True, _fill_value=True
    ):
        """Return the bounds data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `bounds`, `get_bounds`, `get_data`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there are
                no bounds data.

                {{default Exception}}

        :Returns:

            `Data`
                The bounds data.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> x = f.construct('latitude')
        >>> x.get_bounds_data()
        <{{repr}}Data(5, 2): [[-90.0, ..., 90.0]] degrees_north>

        """
        bounds = self.get_bounds(default=None)
        if bounds is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no bounds data"
            )

        data = bounds.get_data(
            default=None, _units=_units, _fill_value=_fill_value
        )
        if data is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no bounds data"
            )

        return data

    @_inplace_enabled(default=False)
    def insert_dimension(self, position, inplace=False):
        """Expand the shape of the data array.

        Inserts a new size 1 axis into the data array. A corresponding
        axis is also inserted into the bounds data array, if present.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `squeeze`, `transpose`

        :Parameters:

            position: `int`, optional
                Specify the position that the new axis will have in
                the data array. By default the new axis has position
                0, the slowest varying position. Negative integers
                counting from the last position are allowed.

                *Parameter example:*
                  ``position=2``

                *Parameter example:*
                  ``position=-1``

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The new construct with expanded data axes. If the
                operation was in-place then `None` is returned.

        **Examples**

        >>> f.shape
        (19, 73, 96)
        >>> f.insert_dimension(position=3).shape
        (96, 73, 19, 1)
        >>> g = f.insert_dimension(position=-1)
        >>> g.shape
        (19, 73, 1, 96)
        >>> f.bounds.shape
        (19, 73, 96, 4)
        >>> g.bounds.shape
        (19, 73, 1, 96, 4)

        """
        c = _inplace_enabled_define_and_cleanup(self)

        # Parse position
        ndim = c.ndim
        if -ndim - 1 <= position < 0:
            position += ndim + 1
        elif not 0 <= position <= ndim:
            raise ValueError(
                f"Can't insert dimension: Invalid position: {position!r}"
            )

        super(PropertiesDataBounds, c).insert_dimension(position, inplace=True)

        # ------------------------------------------------------------
        # Expand the dims of the bounds
        # ------------------------------------------------------------
        bounds = c.get_bounds(None)
        if bounds is not None:
            bounds.insert_dimension(position, inplace=True)

        # ------------------------------------------------------------
        # Expand the dims of the interior_ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.insert_dimension(position, inplace=True)

        return c

    @_inplace_enabled(default=False)
    def persist(self, bounds=True, inplace=False):
        """Persist data into memory.

        {{persist description}}

        **Performance**

        `persist` causes delayed operations to be computed.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `array`, `datetime_array`,
                     `{{package}}.Data.persist`

        :Parameters:

            bounds: `bool`, optional
                If True, the default, then also persist any bounds
                data.

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The construct with persisted data. If the operation
                was in-place then `None` is returned.

        """
        c = _inplace_enabled_define_and_cleanup(self)

        super(PropertiesDataBounds, c).persist(inplace=True)

        # Bounds
        bounds = c.get_bounds(None)
        if bounds is not None:
            bounds.persist(inplace=True)

        # Interior_ring
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.persist(inplace=True)

        return c

    def replace_directory(
        self,
        old=None,
        new=None,
        normalise=False,
        common=False,
    ):
        """Replace file directories in-place.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `file_directories`, `get_filenames`

        :Parameters:

            {{replace old: `str` or `None`, optional}}

            {{replace new: `str` or `None`, optional}}

            {{replace normalise: `bool`, optional}}

            common: `bool`, optional
                If True the base directory structure that is common to
                all files with *new*.

        :Returns:

            `None`

        """
        directory = super().replace_directory(
            old=old, new=new, normalise=normalise, common=common
        )

        bounds = self.get_bounds(None)
        if bounds is not None:
            bounds.replace_directory(
                old=old, new=new, normalise=normalise, common=common
            )

        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.replace_directory(
                old=old, new=new, normalise=normalise, common=common
            )

        return directory

    def nc_clear_hdf5_chunksizes(self, bounds=True, interior_ring=True):
        """Clear the HDF5 chunking strategy for the data.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `nc_hdf5_chunksizes`, `nc_set_hdf5_chunksizes`,
                     `{{package}}.read`, `{{package}}.write`

        :Parameters:

            bounds: `bool`, optional
                If True, the default, then clear the HDF5 chunking
                strategy from any bounds. If False then leave the
                bounds chunking strategy unchanged.

            interior_ring: `bool`
                If True, the default, then clear the HDF5 chunking
                strategy from a geometry interior ring variable. If
                False then leave the geometry interior ring variable
                chunking strategy unchanged.

        :Returns:

            `None` or `str` or `int` or `tuple` of `int`
                The chunking strategy prior to being cleared, as would
                be returned by `nc_hdf5_chunksizes`.

        """
        super().nc_clear_hdf5_chunksizes()

        # Clear the bounds HDF5 chunks
        if bounds:
            bounds = self.get_bounds(None)
            if bounds is not None:
                bounds.nc_clear_hdf5_chunksizes()

        # Clear the  interior_ring HDF5 chunks
        if interior_ring:
            interior_ring = self.get_interior_ring(None)
            if interior_ring is not None:
                interior_ring.nc_clear_hdf5_chunksizes()

    def nc_set_hdf5_chunksizes(
        self, chunksizes, bounds=True, interior_ring=True
    ):
        """Set the HDF5 chunking strategy.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `nc_hdf5_chunksizes`, `nc_clear_hdf5_chunksizes`,
                     `{{package}}.read`, `{{package}}.write`

        :Parameters:

            {{hdf5 chunksizes}}
                  Each dictionary key is an integer that specifies an
                  axis by its position in the data array.

            bounds: `bool`, optional
                If True, the default, then apply the HDF5 chunking
                strategy to the corresponding axes of any bounds. If
                False then leave the bounds chunking strategy
                unchanged.

            interior_ring: `bool`
                If True, the default, then apply the HDF5 chunking
                strategy to the corresponding axis of a geometry
                interior ring variable. If False then leave the
                geometry interior ring variable chunking strategy
                unchanged.

        :Returns:

            `None`

        """
        super().nc_set_hdf5_chunksizes(chunksizes)

        c = self.nc_hdf5_chunksizes()
        if isinstance(c, tuple):
            c = {n: value for n, value in enumerate(c)}

        # Set the bounds HDF5 chunks
        if bounds:
            bounds = self.get_bounds(None)
            if bounds is not None:
                bounds.nc_set_hdf5_chunksizes(c)

        # Set the  interior_ring HDF5 chunks
        if interior_ring:
            interior_ring = self.get_interior_ring(None)
            if interior_ring is not None:
                interior_ring.nc_set_hdf5_chunksizes(c)

    def set_node_count(self, node_count, copy=True):
        """Set the node count variable for geometry bounds.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `del_node_count`, `get_node_count`,
                     `has_node_count`

        :Parameters:

            node_count: `NodeCount`
                The node count variable to be inserted.

            copy: `bool`, optional
                If True (the default) then copy the node count
                variable prior to insertion.

        :Returns:

            `None`

        **Examples**

        >>> n = {{package}}.NodeCount(properties={'long_name': 'node counts'})
        >>> c.set_node_count(n)
        >>> c.has_node_count()
        True
        >>> c.get_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.del_node_count()
        <{{repr}}NodeCount: long_name=node counts>
        >>> c.has_node_count()
        False

        """
        if copy:
            node_count = node_count.copy()

        self._set_component("node_count", node_count, copy=False)

    def set_part_node_count(self, part_node_count, copy=True):
        """Set the part node count variable for geometry bounds.

        .. versionadded:: (cfdm) 1.8.0

        .. seealso:: `del_part_node_count`, `get_part_node_count`,
                     `has_part_node_count`

        :Parameters:

            part_node_count: `PartNodeCount`
                The part node count variable to be inserted.

            copy: `bool`, optional
                If True (the default) then copy the part node count
                variable prior to insertion.

        :Returns:

            `None`

        **Examples**

        >>> p = {{package}}.PartNodeCount(properties={'long_name':
        ...                                           'part node counts'})
        >>> c.set_part_node_count(p)
        >>> c.has_part_node_count()
        True
        >>> c.get_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.del_part_node_count()
        <{{repr}}PartNodeCount: long_name=part node counts>
        >>> c.has_part_node_count()
        False

        """
        if copy:
            part_node_count = part_node_count.copy()

        self._set_component("part_node_count", part_node_count, copy=False)

    @_inplace_enabled(default=False)
    def squeeze(self, axes=None, inplace=False):
        """Remove size one axes from the data array.

        By default all size one axes are removed, but particular size
        one axes may be selected for removal. Corresponding axes are
        also removed from the bounds data array, if present.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `insert_dimension`, `transpose`

        :Parameters:

            axes: (sequence of) `int`, optional
                The positions of the size one axes to be removed. By
                default all size one axes are removed.

                {{axes int examples}}

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The new construct with removed data axes. If the
                operation was in-place then `None` is returned.

        **Examples**

        >>> f.shape
        (1, 73, 1, 96)
        >>> f.squeeze().shape
        (73, 96)
        >>> f.squeeze(0).shape
        (73, 1, 96)
        >>> g = f.squeeze([-3, 2])
        >>> g.shape
        (73, 96)
        >>> f.bounds.shape
        (1, 73, 1, 96, 4)
        >>> g.shape
        (73, 96, 4)

        """
        if axes is None:
            axes = tuple([i for i, n in enumerate(self.shape) if n == 1])

        c = _inplace_enabled_define_and_cleanup(self)
        super(PropertiesDataBounds, c).squeeze(axes, inplace=True)

        # ------------------------------------------------------------
        # Squeeze the bounds
        # ------------------------------------------------------------
        bounds = c.get_bounds(None)
        if bounds is not None:
            bounds.squeeze(axes, inplace=True)

        # ------------------------------------------------------------
        # Squeeze the interior_ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.squeeze(axes, inplace=True)

        return c

    @_inplace_enabled(default=False)
    def to_memory(self, inplace=False):
        """Bring data on disk into memory.

        There is no change to data that is already in memory.

        :Parameters:

            inplace: `bool`, optional
                If True then do the operation in-place and return `None`.

        :Returns:

            `{{class}}` or `None`
                A copy with the data in memory, or `None` if the
                operation was in-place.

        """
        c = _inplace_enabled_define_and_cleanup(self)

        data = c.get_data(None)
        if data is not None:
            data.to_memory(inplace=True)

        bounds = c.get_bounds(None)
        if bounds is not None:
            bounds.to_memory(inplace=True)

        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.to_memory(inplace=True)

        return c

    @_inplace_enabled(default=False)
    def transpose(self, axes=None, inplace=False):
        """Permute the axes of the data array.

        Corresponding axes of the bounds data array, if present, are
        also permuted.

        Note that if i) the data array is two-dimensional, ii) the two
        axes have been permuted, and iii) each cell has four bounds
        values; then columns 1 and 3 (counting from 0) of the bounds
        axis are swapped to preserve contiguity bounds in adjacent
        cells. See section 7.1 "Cell Boundaries" of the CF conventions
        for details.

        .. seealso:: `insert_dimension`, `squeeze`

        :Parameters:

            axes: (sequence of) `int`, optional
                The new axis order. By default the order is reversed.

                {{axes int examples}}

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The new construct with permuted data axes. If the
                operation was in-place then `None` is returned.

        **Examples**

        >>> f.shape
        (19, 73, 96)
        >>> f.transpose().shape
        (96, 73, 19)
        >>> g = f.transpose([1, 0, 2])
        >>> g.shape
        (73, 19, 96)
        >>> f.bounds.shape
        (19, 73, 96, 4)
        >>> g.bounds.shape
        (73, 19, 96, 4)

        """
        ndim = self.ndim
        if axes is None:
            axes = list(range(ndim - 1, -1, -1))
        else:
            axes = self._parse_axes(axes)

        # ------------------------------------------------------------
        # Transpose the coordinates
        # ------------------------------------------------------------
        c = _inplace_enabled_define_and_cleanup(self)
        super(PropertiesDataBounds, c).transpose(axes, inplace=True)

        # ------------------------------------------------------------
        # Transpose the bounds
        # ------------------------------------------------------------
        bounds = c.get_bounds(None)
        data = c.get_bounds_data(None, _fill_value=False)
        if bounds is not None:
            b_axes = axes[:]
            b_axes.extend(list(range(ndim, data.ndim)))
            bounds.transpose(b_axes, inplace=True)

            data = bounds.get_data(None, _fill_value=False)
            if (
                data is not None
                and ndim == 2
                and data.ndim == 3
                and data.shape[-1] == 4
                and b_axes[0:2] == [1, 0]
            ):
                # Swap elements 1 and 3 of the trailing dimension so
                # that the values are still contiguous (if they ever
                # were). See section 7.1 of the CF conventions.
                data[:, :, slice(1, 4, 2)] = data[:, :, slice(3, 0, -2)]
                bounds.set_data(data, copy=False)

        # ------------------------------------------------------------
        # Transpose the interior ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.transpose(axes + [-1], inplace=True)

        return c

    @_inplace_enabled(default=False)
    def uncompress(self, inplace=False):
        """Uncompress the construct.

        Compression saves space by identifying and removing unwanted
        missing data. Such compression techniques store the data more
        efficiently and result in no precision loss.

        Whether or not the construct is compressed does not alter its
        functionality nor external appearance.

        A construct that is already uncompressed will be returned
        unchanged.

        The following type of compression are available:

            * Ragged arrays for discrete sampling geometries
              (DSG). Three different types of ragged array
              representation are supported.

            ..

            * Compression by gathering.

        .. versionadded:: (cfdm) 1.7.11

        :Parameters:

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The uncompressed construct, or `None` if the operation
                was in-place.

        **Examples**

        >>> f.data.get_compression_type()
        'ragged contiguous'
        >>> g = f.uncompress()
        >>> g.data.get_compression_type()
        ''
        >>> g.equals(f)
        True

        """
        v = _inplace_enabled_define_and_cleanup(self)
        super(PropertiesDataBounds, v).uncompress(inplace=True)

        bounds = v.get_bounds(None)
        if bounds is not None:
            bounds.uncompress(inplace=True)

        return v



import logging

import h5netcdf

from . import abstract
from .locks import netcdf_lock
from .mixin import IndexMixin
from .netcdfindexer import netcdf_indexer

logger = logging.getLogger(__name__)


class H5netcdfArray(IndexMixin, abstract.FileArray):
    """A netCDF array accessed with `h5netcdf`.

    .. versionadded:: (cfdm) 1.11.2.0

    """

    @property
    def _lock(self):
        """Return the lock used for netCDF file access.

        Returns a lock object that prevents concurrent reads of netCDF
        files, which are not currently supported by `h5netcdf`.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return netcdf_lock

    def _get_array(self, index=None):
        """Returns a subspace of the dataset variable.

        The subspace is defined by the `index` attributes, and is
        applied with `cfdm.netcdf_indexer`.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `__array__`, `index`

        :Parameters:

            {{index: `tuple` or `None`, optional}}

        :Returns:

            `numpy.ndarray`
                The subspace.

        """
        if index is None:
            index = self.index()

        # Note: We need to lock because HDF5 is about to access the
        #       file.
        with self._lock:
            dataset, address = self.open()
            dataset0 = dataset

            groups, address = self.get_groups(address)
            if groups:
                dataset = self._group(dataset, groups)

            # Get the variable by netCDF name
            variable = dataset.variables[address]

            # Get the data, applying masking and scaling as required.
            array = netcdf_indexer(
                variable,
                mask=self.get_mask(),
                unpack=self.get_unpack(),
                always_masked_array=False,
                orthogonal_indexing=True,
                copy=False,
            )
            array = array[index]

            # Set the attributes, if they haven't been set already.
            self._set_attributes(variable)

            self.close(dataset0)
            del dataset, dataset0

        return array

    def _group(self, dataset, groups):
        """Return the group object containing a variable.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            dataset: `h5netcdf.File`
                The dataset containing the variable.

            groups: sequence of `str`
                The definition of which group the variable is in. For
                instance, if the variable is in group
                ``/forecast/model`` then *groups* would be
                ``['forecast', 'model']``.

        :Returns:

            `h5netcdf.File` or `h5netcdf.Group`
                The group object, which might be the root group.

        """
        for g in groups:
            dataset = dataset.groups[g]

        return dataset

    def _set_attributes(self, var):
        """Set the netCDF variable attributes.

        These are set from the netCDF variable attributes, but only if
        they have not already been defined, either during `{{class}}`
        instantiation or by a previous call to `_set_attributes`.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            var: `h5netcdf.Variable`
                The netCDF variable.

        :Returns:

            `None`

        """
        if self._get_component("attributes", None) is not None:
            return

        self._set_component("attributes", dict(var.attrs), copy=False)

    def close(self, dataset):
        """Close the dataset containing the data.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            dataset: `h5netcdf.File`
                The netCDF dataset to be closed.

        :Returns:

            `None`

        """
        if self._get_component("close"):
            dataset.close()

    def get_groups(self, address):
        """The netCDF4 group structure of a netCDF variable.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            address: `str` or `int`
                The netCDF variable name, or integer varid, from which
                to get the groups.

        :Returns:

            (`list`, `str`) or (`list`, `int`)
                The group structure and the name within the group. If
                *address* is a varid then an empty list and the varid
                are returned.

        **Examples**

        >>> n.get_groups('tas')
        ([], 'tas')

        >>> n.get_groups('/tas')
        ([], 'tas')

        >>> n.get_groups('/data/model/tas')
        (['data', 'model'], 'tas')

        >>> n.get_groups(9)
        ([], 9)

        """
        try:
            if "/" not in address:
                return [], address
        except TypeError:
            return [], address

        out = address.split("/")[1:]
        return out[:-1], out[-1]

    def open(self, **kwargs):
        """Return a dataset file object and address.

        When multiple files have been provided an attempt is made to
        open each one, in the order stored, and a file object is
        returned from the first file that exists.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            (`h5netcdf.File`, `str`)
                The open file object, and the address of the data
                within the file.

        """
        return super().open(
            h5netcdf.File, mode="r", decode_vlen_strings=True, **kwargs
        )



import datetime
import faulthandler
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

# Create test cell connectivity object
c = cfdm.CellConnectivity()
c.set_properties({"long_name": "neighbour faces for faces"})
c.nc_set_variable("Mesh2_face_links")
data = cfdm.Data(
    [
        [0, 1, 2, -99, -99],
        [1, 0, -99, -99, -99],
        [2, 0, -99, -99, -99],
    ],
    dtype="i4",
)
data.masked_values(-99, inplace=True)
c.set_data(data)
c.set_connectivity("edge")


class CellConnectivityTest(unittest.TestCase):
    """Unit test for the CellConnectivity class."""

    c = c

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_CellConnectivity__repr__str__dump(self):
        """Test all means of CellConnectivity inspection."""
        c = self.c
        self.assertEqual(
            repr(c), "<CellConnectivity: connectivity:edge(3, 5) >"
        )
        self.assertEqual(str(c), "connectivity:edge(3, 5) ")
        self.assertEqual(
            c.dump(display=False),
            """Cell Connectivity: connectivity:edge
    long_name = 'neighbour faces for faces'
    Data(3, 5) = [[0, ..., --]]""",
        )

    def test_CellConnectivity_copy(self):
        """Test the copy of CellConnectivity."""
        c = self.c
        self.assertTrue(c.equals(c.copy()))

    def test_CellConnectivity_data(self):
        """Test the data of CellConnectivity."""
        c = self.c
        self.assertEqual(c.ndim, 1)

    def test_CellConnectivity_connectivity(self):
        """Test the 'connectivity' methods of CellConnectivity."""
        c = self.c.copy()
        self.assertTrue(c.has_connectivity())
        self.assertEqual(c.get_connectivity(), "edge")
        self.assertEqual(c.del_connectivity(), "edge")
        self.assertFalse(c.has_connectivity())
        self.assertIsNone(c.get_connectivity(None))
        self.assertIsNone(c.del_connectivity(None))

        with self.assertRaises(ValueError):
            c.get_connectivity()

        with self.assertRaises(ValueError):
            c.del_connectivity()

        self.assertIsNone(c.set_connectivity("edge"))
        self.assertTrue(c.has_connectivity())
        self.assertEqual(c.get_connectivity(), "edge")

    def test_CellConnectivity_transpose(self):
        """Test the 'transpose' method of CellConnectivity."""
        c = self.c.copy()
        d = c.transpose()
        self.assertTrue(c.equals(d))
        self.assertIsNone(c.transpose(inplace=True))

        for axes in ([1], [1, 0], [3]):
            with self.assertRaises(ValueError):
                c.transpose(axes)

    def test_CellConnectivity_normalise(self):
        """Test the 'normalise' method of CellConnectivity."""
        data = cfdm.Data(
            [[4, 1, 10, 125], [1, 4, -99, -99], [125, 4, -99, -99]],
            mask_value=-99,
        )
        d = cfdm.CellConnectivity(connectivity="edge", data=data)

        n = np.ma.array([[0, 1, 2, -99], [1, 0, -99, -99], [2, 0, -99, -99]])
        n = np.ma.where(n == -99, np.ma.masked, n)

        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())
        d1 = d.normalise(start_index=1).array
        self.assertEqual(d1.shape, n.shape)
        self.assertTrue((d1.mask == n.mask).all())
        self.assertTrue((d1 == n + 1).all())

        e = d.copy()
        self.assertIsNone(e.normalise(inplace=True))
        d0 = e.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())
        d1 = e.normalise(start_index=1).array
        self.assertEqual(d1.shape, n.shape)
        self.assertTrue((d1.mask == n.mask).all())
        self.assertTrue((d1 == n + 1).all())

        self.assertIsNone(e.normalise(start_index=1, inplace=True))
        d0 = e.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())
        d1 = e.normalise(start_index=1).array
        self.assertEqual(d1.shape, n.shape)
        self.assertTrue((d1.mask == n.mask).all())
        self.assertTrue((d1 == n + 1).all())

        d.data[:, -1] = np.ma.masked
        n = np.ma.array([[0, 1, -99, -99], [1, 0, -99, -99], [2, 0, -99, -99]])
        n = np.ma.where(n == -99, np.ma.masked, n)

        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())

        d.data[:, 1:] = np.ma.masked
        n = np.ma.array(
            [[0, -99, -99, -99], [1, -99, -99, -99], [2, -99, -99, -99]]
        )
        n = np.ma.where(n == -99, np.ma.masked, n)
        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())

        n = np.ma.array([[0], [1], [2]])
        d0 = d.normalise(remove_empty_columns=True).array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_dsg.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tempfile,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class DSGTest(unittest.TestCase):
    """Test discrete sampling geometry ragged array representations."""

    contiguous = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "DSG_timeSeries_contiguous.nc",
    )
    indexed = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "DSG_timeSeries_indexed.nc"
    )
    indexed_contiguous = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "DSG_timeSeriesProfile_indexed_contiguous.nc",
    )

    c = cfdm.read(contiguous)
    i = cfdm.read(indexed)
    ic = cfdm.read(indexed_contiguous)

    a = np.ma.masked_all((4, 9), dtype=float)
    a[0, 0:3] = [0.0, 1.0, 2.0]
    a[1, 0:7] = [1.0, 11.0, 21.0, 31.0, 41.0, 51.0, 61.0]
    a[2, 0:5] = [2.0, 102.0, 202.0, 302.0, 402.0]
    a[3, 0:9] = [
        3.0,
        1003.0,
        2003.0,
        3003.0,
        4003.0,
        5003.0,
        6003.0,
        7003.0,
        8003.0,
    ]

    b = np.array(
        [
            [
                [20.7, -99, -99, -99],
                [10.1, 11.8, 18.2, -99],
                [11.0, 11.8, -99, -99],
                [16.3, 20.0, -99, -99],
                [13.8, 18.3, -99, -99],
                [15.9, -99, -99, -99],
                [15.7, 21.2, -99, -99],
                [22.5, -99, -99, -99],
                [18.0, -99, -99, -99],
                [12.6, 21.7, -99, -99],
                [10.5, 12.9, 21.0, -99],
                [16.0, 19.7, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
            ],
            [
                [5.2, 5.8, 10.8, 13.8],
                [2.6, 9.2, -99, -99],
                [0.7, 4.0, -99, -99],
                [15.7, -99, -99, -99],
                [2.5, 16.0, -99, -99],
                [4.6, 9.8, -99, -99],
                [0.6, 3.1, -99, -99],
                [3.8, -99, -99, -99],
                [5.7, 12.9, 18.1, -99],
                [3.9, 6.9, 16.9, -99],
                [7.3, 13.8, 16.0, -99],
                [4.5, 9.8, 11.3, -99],
                [1.5, -99, -99, -99],
                [0.9, 4.3, 6.2, -99],
                [1.7, 9.9, -99, -99],
                [9.3, -99, -99, -99],
                [0.7, -99, -99, -99],
                [15.7, -99, -99, -99],
                [0.7, 1.2, -99, -99],
                [4.5, 12.4, 13.0, -99],
                [3.5, 6.8, 7.9, -99],
                [8.1, 12.2, -99, -99],
                [5.9, -99, -99, -99],
                [1.0, 9.6, -99, -99],
                [5.6, 7.8, 9.1, -99],
                [7.1, 9.0, 10.4, -99],
            ],
            [
                [35.2, -99, -99, -99],
                [34.7, 38.9, 48.1, -99],
                [35.2, 39.3, 39.6, -99],
                [40.3, 40.4, 48.0, -99],
                [30.0, 36.5, -99, -99],
                [33.3, 43.3, -99, -99],
                [37.7, -99, -99, -99],
                [33.5, -99, -99, -99],
                [31.9, 33.7, -99, -99],
                [34.1, 35.4, 41.0, -99],
                [30.2, 33.7, 38.7, -99],
                [32.4, 42.4, -99, -99],
                [33.2, 34.9, 39.7, -99],
                [33.2, -99, -99, -99],
                [38.5, -99, -99, -99],
                [37.3, 39.9, -99, -99],
                [30.0, 39.1, -99, -99],
                [36.4, 39.1, 45.6, -99],
                [41.0, -99, -99, -99],
                [31.1, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
                [-99, -99, -99, -99],
            ],
        ]
    )

    b = np.ma.where(b == -99, np.ma.masked, b)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_DSG_contiguous(self):
        """Test the contiguous ragged array DSG representation."""
        f = self.c.copy()

        self.assertEqual(len(f), 2)

        # Select the specific humidity field
        q = [
            g
            for g in f
            if g.get_property("standard_name") == "specific_humidity"
        ][0]

        self.assertTrue(q._equals(self.a, q.data.array))

        cfdm.write(f, tempfile)
        g = cfdm.read(tempfile)

        self.assertEqual(len(g), len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], verbose=3))

        # ------------------------------------------------------------
        # Test creation
        # ------------------------------------------------------------
        # Define the ragged array values
        ragged_array = np.array(
            [280, 282.5, 281, 279, 278, 279.5], dtype="float32"
        )

        # Define the count array values
        count_array = [2, 4]

        # Create the count variable
        count_variable = cfdm.Count(data=cfdm.Data(count_array))
        count_variable.set_property(
            "long_name", "number of obs for this timeseries"
        )

        # Create the contiguous ragged array object
        array = cfdm.RaggedContiguousArray(
            compressed_array=ragged_array,
            shape=(2, 4),
            size=8,
            ndim=2,
            count_variable=count_variable,
        )

        # Create the field construct with the domain axes and the ragged
        # array
        tas = cfdm.Field()
        tas.set_properties(
            {
                "standard_name": "air_temperature",
                "units": "K",
                "featureType": "timeSeries",
            }
        )

        # Create the domain axis constructs for the uncompressed array
        X = tas.set_construct(cfdm.DomainAxis(4))
        Y = tas.set_construct(cfdm.DomainAxis(2))

        # Set the data for the field
        tas.set_data(cfdm.Data(array), axes=[Y, X])

        cfdm.write(tas, tempfile)

    def test_DSG_indexed(self):
        """Test the indexed ragged array DSG representation."""
        f = self.i.copy()

        self.assertEqual(len(f), 2)

        # Select the specific humidity field
        q = [
            g
            for g in f
            if g.get_property("standard_name") == "specific_humidity"
        ][0]

        self.assertTrue(q._equals(q.data.array, self.a))

        cfdm.write(f, tempfile)
        g = cfdm.read(tempfile)

        self.assertEqual(len(g), len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], verbose=3))

    def test_DSG_indexed_contiguous(self):
        """Test the indexed contiguous ragged array representation."""
        f = self.ic.copy()

        self.assertEqual(len(f), 2)

        # Select the specific humidity field
        q = [
            g
            for g in f
            if g.get_property("standard_name") == "specific_humidity"
        ][0]

        qa = q.data.array

        self.assertTrue(q._equals(qa, self.b))

        cfdm.write(f, tempfile)
        g = cfdm.read(tempfile)

        self.assertEqual(len(g), len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], verbose=3))

        # Check original filenames
        self.assertEqual(
            q.get_original_filenames(), set([self.indexed_contiguous])
        )

    def test_DSG_create_contiguous(self):
        """Test the creation of a contiguous ragged array."""
        # Define the ragged array values
        ragged_array = np.array([1, 3, 4, 3, 6], dtype="float32")
        # Define the count array values
        count_array = [2, 3]

        # Initialise the count variable
        count_variable = cfdm.Count(data=cfdm.Data(count_array))
        count_variable.set_property(
            "long_name", "number of obs for this timeseries"
        )

        # Initialise the contiguous ragged array object
        array = cfdm.RaggedContiguousArray(
            compressed_array=cfdm.Data(ragged_array),
            shape=(2, 3),
            size=6,
            ndim=2,
            count_variable=count_variable,
        )

        # Initialize the auxiliary coordinate construct with the
        # ragged array and set some properties
        z = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(array),
            properties={
                "standard_name": "height",
                "units": "km",
                "positive": "up",
            },
        )

        self.assertTrue(
            (
                z.data.array
                == np.ma.masked_array(
                    data=[[1.0, 3.0, 99], [4.0, 3.0, 6.0]],
                    mask=[[False, False, True], [False, False, False]],
                    fill_value=1e20,
                    dtype="float32",
                )
            ).all()
        )

        self.assertEqual(z.data.get_compression_type(), "ragged contiguous")

        self.assertTrue(
            (
                z.data.compressed_array
                == np.array([1.0, 3.0, 4.0, 3.0, 6.0], dtype="float32")
            ).all()
        )

        self.assertTrue(
            (z.data.get_count().data.array == np.array([2, 3])).all()
        )


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



import numpy as np

from .. import abstract
from .abstract import Array
from .numpyarray import NumpyArray


class Data(abstract.Container):
    """An orthogonal multidimensional array with masking and units.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        array=None,
        units=None,
        calendar=None,
        fill_value=None,
        source=None,
        copy=True,
        _use_array=True,
        _custom_deep_copy=False,
    ):
        """**Initialisation**

        :Parameters:

            array: data_like, optional
                The array of values.

                {{data_like}}

                Note that for `{{class}}` objects and {{package}}
                objects that contain `{{class}}` objects,
                ``{{package}}.{{class}}(array)`` is equivalent to
                ``{{package}}.{{class}}(source=array.__data__())``.

            array: `numpy` array_like, optional
                The array of values.

                Note that if *array* is another `{{class}}` instance
                then ``{{package}}.{{class}}(array)`` is equivalent to
                ``{{package}}.{{class}}(source=array)``.

            units: `str`, optional
                The physical units of the data.

                The units may also be set after initialisation with
                the `set_units` method.

                *Parameter example:*
                  ``units='km hr-1'``

                *Parameter example:*
                  ``units='days since 2018-12-01'``

            calendar: `str`, optional
                The calendar for reference time units.

                The calendar may also be set after initialisation with
                the `set_calendar` method.

                *Parameter example:*
                  ``calendar='360_day'``

            fill_value: optional
                The fill value of the data. By default, or if set to
                `None`, the `numpy` fill value appropriate to the
                array's data type will be used (see
                `numpy.ma.default_fill_value`).

                The fill value may also be set after initialisation
                with the `set_fill_value` method.

                *Parameter example:*
                  ``fill_value=-999.``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        if source is None:
            try:
                source = array.__data__()
            except (AttributeError, ValueError, TypeError):
                pass

        super().__init__(source=source)

        if source is not None:
            try:
                array = source._get_Array(None)
            except AttributeError:
                array = None

            try:
                units = source.get_units(None)
            except AttributeError:
                units = None

            try:
                calendar = source.get_calendar(None)
            except AttributeError:
                calendar = None

            try:
                fill_value = source.get_fill_value(None)
            except AttributeError:
                fill_value = None

        if units is not None:
            self._set_component("units", units, copy=False)

        if calendar is not None:
            self._set_component("calendar", calendar, copy=False)

        if fill_value is not None:
            self._set_component("fill_value", fill_value, copy=False)

        if _use_array and array is not None:
            self._set_Array(array, copy=copy)

    def __data__(self):
        """Defines the data interface.

        Returns a new reference to the data.

        :Returns:

            `{{class}}`

        """
        return self

    @property
    def array(self):
        """Return an independent `numpy` array containing the data.

        If a fill value has been set (see `set_fill_value`) then it
        will be used, otherwise the default numpy fill value
        appropriate to the data type will be used.

        :Returns:

            `numpy.ndarray`
                An independent `numpy` array of the data.

        **Examples**

        >>> d = {{package}}.{{class}}([1, 2, 3.0], 'km')
        >>> n = d.array
        >>> isinstance(n, numpy.ndarray)
        True
        >>> print(n)
        [1. 2. 3.]
        >>> n[0] = 88
        >>> d
        <{{repr}}{{class}}(3): [1.0, 2.0, 3.0] km>

        """
        array = self._get_Array().array

        # Set the numpy array fill value
        if np.ma.isMA(array):
            array.set_fill_value(self.get_fill_value(None))

        return array

    @property
    def dtype(self):
        """Data-type of the data elements.

        **Examples**

        >>> d.dtype
        dtype('float64')
        >>> type(d.dtype)
        <type 'numpy.dtype'>

        """
        datatype = self._get_Array().dtype
        if datatype is None:
            # The datatype is not known, so get the numpy array and
            # get it from that.
            datatype = self.array.dtype

        return datatype

    @property
    def ndim(self):
        """Number of data dimensions.

        **Examples**

        >>> d.shape
        (73, 96)
        >>> d.ndim
        2
        >>> d.size
        7008

        >>> d.shape
        (1, 1, 1)
        >>> d.ndim
        3
        >>> d.size
        1

        >>> d.shape
        ()
        >>> d.ndim
        0
        >>> d.size
        1

        """
        return self._get_Array().ndim

    @property
    def shape(self):
        """Tuple of data dimension sizes.

        **Examples**

        >>> d.shape
        (73, 96)
        >>> d.ndim
        2
        >>> d.size
        7008

        >>> d.shape
        (1, 1, 1)
        >>> d.ndim
        3
        >>> d.size
        1

        >>> d.shape
        ()
        >>> d.ndim
        0
        >>> d.size
        1

        """
        return self._get_Array().shape

    @property
    def size(self):
        """Number of elements in the data.

        **Examples**

        >>> d.shape
        (73, 96)
        >>> d.size
        7008
        >>> d.ndim
        2

        >>> d.shape
        (1, 1, 1)
        >>> d.ndim
        3
        >>> d.size
        1

        >>> d.shape
        ()
        >>> d.ndim
        0
        >>> d.size
        1

        """
        return self._get_Array().size

    def copy(self, array=True):
        """Return a deep copy of the data.

        ``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

        Copy-on-write is employed, so care must be taken when modifying
        any attribute.

        :Parameters:

            array: `bool`, optional
                If True (the default) then copy the array, else the
                array is not copied.

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> e = d.copy()
        >>> e = d.copy(array=False)

        """
        return type(self)(source=self, copy=True, _use_array=array)

    def _del_Array(self, default=ValueError()):
        """Delete the underlying array.

        :Returns:

                The array.

        **Examples**

        >>> old = d.del_data()

        """
        return self._del_component("array", default=default)

    def del_calendar(self, default=ValueError()):
        """Delete the calendar.

        .. seealso:: `get_calendar`, `has_calendar`, `set_calendar`,
                     `del_units`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                calendar has not been set.

                {{default Exception}}

        :Returns:

            `str`
                The value of the deleted calendar.

        **Examples**

        >>> d.set_calendar('360_day')
        >>> d.has_calendar()
        True
        >>> d.get_calendar()
        '360_day'
        >>> d.del_calendar()
        >>> d.has_calendar()
        False
        >>> d.get_calendar()
        ValueError: Can't get non-existent calendar
        >>> print(d.get_calendar(None))
        None
        >>> print(d.del_calendar(None))
        None

        """
        return self._del_component("calendar", default)

    def del_fill_value(self, default=ValueError()):
        """Delete the fill value.

        .. seealso:: `get_fill_value`, `has_fill_value`, `set_fill_value`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the fill
                value has not been set.

                {{default Exception}}

        :Returns:

                The value of the deleted fill value.

        **Examples**

        >>> f.set_fill_value(-9999)
        >>> f.has_fill_value()
        True
        >>> f.get_fill_value()
        -9999
        >>> print(f.del_fill_value())
        -9999
        >>> f.get_fill_value()
        ValueError: Can't get non-existent fill value
        >>> f.get_fill_value(10**10)
        10000000000
        >>> print(f.get_fill_value(None))
        None
        >>> f.set_fill_value(None)
        >>> print(f.get_fill_value())
        None
        >>> f.has_fill_value()
        False

        """
        return self._del_component("fill_value", default)

    def del_units(self, default=ValueError()):
        """Delete the units.

        .. seealso:: `get_units`, `has_units`, `set_units`, `del_calendar`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the units
                has not been set.

                {{default Exception}}

        :Returns:

            `str`
                The value of the deleted units.

        **Examples**

        >>> d.set_units('metres')
        >>> d.has_units()
        True
        >>> d.get_units()
        'metres'
        >>> d.del_units()
        >>> d.has_units()
        False
        >>> d.get_units()
        ValueError: Can't get non-existent units
        >>> print(d.get_units(None))
        None
        >>> print(d.del_units(None))
        None

        """
        return self._del_component("units", default)

    def get_calendar(self, default=ValueError()):
        """Return the calendar.

        .. seealso:: `del_calendar`, `has_calendar`, `set_calendar`,
                     `get_units`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                calendar has not been set.

                {{default Exception}}

        :Returns:

            `str`
                The calendar.

        **Examples**

        >>> d.set_calendar('360_day')
        >>> d.has_calendar()
        True
        >>> d.get_calendar()
        '360_day'
        >>> d.del_calendar()
        >>> d.has_calendar()
        False
        >>> d.get_calendar()
        ValueError: Can't get non-existent calendar
        >>> print(d.get_calendar(None))
        None
        >>> print(d.del_calendar(None))
        None

        """
        return self._get_component("calendar", default)

    def _get_Array(self, default=ValueError()):
        """Return the array object.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the array
                has not been set

                {{default Exception}}

        :Returns:

                The array object.

        **Examples**

        >>> a = d._get_Array(None)

        """
        return self._get_component("array", default)

    def get_fill_value(self, default=ValueError()):
        """Return the missing data value.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_fill_value`, `has_fill_value`, `set_fill_value`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the fill
                value has not been set.

                {{default Exception}}

        :Returns:

                The fill value.

        **Examples**

        >>> f.set_fill_value(-9999)
        >>> f.has_fill_value()
        True
        >>> f.get_fill_value()
        -9999
        >>> print(f.del_fill_value())
        -9999
        >>> f.get_fill_value()
        ValueError: Can't get non-existent fill value
        >>> f.get_fill_value(10**10)
        10000000000
        >>> print(f.get_fill_value(None))
        None
        >>> f.set_fill_value(None)
        >>> print(f.get_fill_value())
        None
        >>> f.has_fill_value()
        False

        """
        return self._get_component("fill_value", default)

    def get_units(self, default=ValueError()):
        """Return the units.

        .. seealso:: `del_units`, `has_units`, `set_units`, `get_calendar`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the units
                has not been set.

                {{default Exception}}

        :Returns:

            `str`
                The units.

        **Examples**

        >>> d.set_units('metres')
        >>> d.has_units()
        True
        >>> d.get_units()
        'metres'
        >>> d.del_units()
        >>> d.has_units()
        False
        >>> d.get_units()
        ValueError: Can't get non-existent units
        >>> print(d.get_units(None))
        None
        >>> print(d.del_units(None))
        None

        """
        return self._get_component("units", default)

    def has_units(self):
        """Whether units have been set.

        .. seealso:: `del_units`, `get_units`, `set_units`, `has_calendar`

        :Returns:

            `bool`
                True if units have been set, otherwise False.

        **Examples**

        >>> d.set_units('metres')
        >>> d.has_units()
        True
        >>> d.get_units()
        'metres'
        >>> d.del_units()
        >>> d.has_units()
        False
        >>> d.get_units()
        ValueError: Can't get non-existent units
        >>> print(d.get_units(None))
        None
        >>> print(d.del_units(None))
        None

        """
        return self._has_component("units")

    def has_calendar(self):
        """Whether a calendar has been set.

        .. seealso:: `del_calendar`, `get_calendar`, `set_calendar`,
                     `has_units`

        :Returns:

            `bool`
                True if the calendar has been set, otherwise False.

        **Examples**

        >>> d.set_calendar('360_day')
        >>> d.has_calendar()
        True
        >>> d.get_calendar()
        '360_day'
        >>> d.del_calendar()
        >>> d.has_calendar()
        False
        >>> d.get_calendar()
        ValueError: Can't get non-existent calendar
        >>> print(d.get_calendar(None))
        None
        >>> print(d.del_calendar(None))
        None

        """
        return self._has_component("calendar")

    def has_fill_value(self):
        """Whether a fill value has been set.

        .. seealso:: `del_fill_value`, `get_fill_value`, `set_fill_value`

        :Returns:

            `bool`
                True if a fill value has been set, otherwise False.

        **Examples**

        >>> f.set_fill_value(-9999)
        >>> f.has_fill_value()
        True
        >>> f.get_fill_value()
        -9999
        >>> print(f.del_fill_value())
        -9999
        >>> f.get_fill_value()
        ValueError: Can't get non-existent fill value
        >>> f.get_fill_value(10**10)
        10000000000
        >>> print(f.get_fill_value(None))
        None
        >>> f.set_fill_value(None)
        >>> print(f.get_fill_value())
        None
        >>> f.has_fill_value()
        False

        """
        return self._has_component("fill_value")

    def set_calendar(self, value):
        """Set the calendar.

        .. seealso:: `del_calendar`, `get_calendar`, `has_calendar`,
                     `set_units`

        :Parameters:

            value: `str`
                The new calendar.

        :Returns:

            `None`

        **Examples**

        >>> d.set_calendar('360_day')
        >>> d.has_calendar()
        True
        >>> d.get_calendar()
        '360_day'
        >>> d.del_calendar()
        >>> d.has_calendar()
        False
        >>> d.get_calendar()
        ValueError: Can't get non-existent calendar
        >>> print(d.get_calendar(None))
        None
        >>> print(d.del_calendar(None))
        None

        """
        return self._set_component("calendar", value, copy=False)

    def _set_Array(self, array, copy=True):
        """Set the array.

        :Parameters:

            array: `numpy` array_like or `Array`, optional
                The array to be inserted.

        :Returns:

            `None`

        **Examples**

        >>> d._set_Array(a)

        """
        if not isinstance(array, Array):
            if not isinstance(array, np.ndarray):
                array = np.asanyarray(array)

            array = NumpyArray(array)

        if copy:
            array = array.copy()

        self._set_component("array", array, copy=False)

    def set_fill_value(self, value):
        """Set the missing data value.

        .. seealso:: `del_fill_value`, `has_fill_value`, `get_fill_value`

        :Parameters:

            value: scalar
                The new fill value.

        :Returns:

            `None`

        **Examples**

        >>> f.set_fill_value(-9999)
        >>> f.has_fill_value()
        True
        >>> f.get_fill_value()
        -9999
        >>> print(f.del_fill_value())
        -9999
        >>> f.get_fill_value()
        ValueError: Can't get non-existent fill value
        >>> f.get_fill_value(10**10)
        10000000000
        >>> print(f.get_fill_value(None))
        None
        >>> f.set_fill_value(None)
        >>> print(f.get_fill_value())
        None
        >>> f.has_fill_value()
        False

        """
        self._set_component("fill_value", value, copy=False)

    def set_units(self, value):
        """Set the units.

        .. seealso:: `del_units`, `get_units`, `has_units`, `set_calendar`

        :Parameters:

            value: `str`
                The new units.

        :Returns:

            `None`

        **Examples**

        >>> d.set_units('metres')
        >>> d.has_units()
        True
        >>> d.get_units()
        'metres'
        >>> d.del_units()
        >>> d.has_units()
        False
        >>> d.get_units()
        ValueError: Can't get non-existent units
        >>> print(d.get_units(None))
        None
        >>> print(d.del_units(None))
        None

        """
        self._set_component("units", value, copy=False)

    def source(self, default=ValueError()):
        """Return the underlying array object.

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the array
                has not been set.

                {{default Exception}}

        :Returns:

            subclass of `Array`
                The underlying array object.

        **Examples**

        >>> f = {{package}}.read('file.nc')[0]
        >>> d = f.data
        >>> d.source()
        <{{repr}}NetCDF4Array(149, 182): file=file.nc variable=latitude>

        """
        return self._get_component("array", default=default)



from .subarray import Subarray


class MeshSubarray(Subarray):
    """A subarray of an array defined by a UGRID connectivity variable.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def __init__(
        self,
        data=None,
        indices=None,
        shape=None,
        compressed_dimensions=None,
        dtype=None,
        start_index=None,
        cell_dimension=None,
        source=None,
        copy=True,
        context_manager=None,
    ):
        """**Initialisation**

        :Parameters:

            data: array_like
                A 2-d integer array that contains zero-based indices
                that identifies UGRID nodes for each cell, as found in
                a UGRID connectivity variable. This array contains the
                indices for all subarrays.

            indices: `tuple` of `slice`
                The indices of *data* that define this subarray.

            shape: `tuple` of `int`
                The shape of the uncompressed subarray.

            {{init compressed_dimensions: `dict`}}

            {{init start_index: `int`}}

            {{init cell_dimension: `int`}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

            context_manager: function, optional
                A context manager that provides a runtime context for
                the conversion of *data* to a `numpy` array.

        """
        super().__init__(
            data=data,
            indices=indices,
            shape=shape,
            compressed_dimensions=compressed_dimensions,
            source=source,
            copy=copy,
            context_manager=context_manager,
        )

        if source is not None:
            try:
                start_index = source.get_start_index(None)
            except AttributeError:
                start_index = None

            try:
                cell_dimension = source.get_cell_dimension(None)
            except AttributeError:
                cell_dimension = None

        if start_index is not None:
            self._set_component("start_index", start_index, copy=False)

        if cell_dimension is not None:
            self._set_component("cell_dimension", cell_dimension, copy=False)

    def _select_data(self, data=None, check_mask=True):
        """Select compressed elements that correspond to this subarray.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            data: array_like or `None`
                A full compressed array spanning all subarrays, from
                which the elements for this subarray will be
                returned. By default, or if `None`, then the `data`
                array is used.

            check_mask: `bool`, optional
                Check for masked elements in the selected data, and if
                there are none convert the output to a non-masked
                `numpy` array.

        :Returns:

            `numpy.ndarray`
                Values of the compressed array that correspond to this
                subarray.

        """
        if data is None:
            data = self.data

        indices = self.indices
        cell_dimension = self.cell_dimension
        if cell_dimension == 1:
            # The 2-d data is stored with the wrong axis order
            indices = indices[::-1]

        array = self._asanyarray(data, indices=indices, check_mask=check_mask)

        if cell_dimension == 1:
            array = array.T

        return array

    @property
    def cell_dimension(self):
        """The position of the dimension that indexes the cells.

        Either ``0`` or ``1``.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self._get_component("cell_dimension")

    @property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self.data.dtype

    @property
    def start_index(self):
        """The start index of values in the compressed data.

        Either ``0`` or ``1``.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self._get_component("start_index")



from .abstract import MeshArray
from .mixin import CompressedArrayMixin
from .subarray import CellConnectivitySubarray


class CellConnectivityArray(CompressedArrayMixin, MeshArray):
    """A connectivity array derived from a UGRID connectivity variable.

    A UGRID connectivity variable contains indices which map each cell
    to its neighbours, as found in a UGRID "face_face_connectivity" or
    "volume_volume_connectivity" variable.

    The connectivity array has one more column than the corresponding
    UGRID variable. The extra column, in the first position, contains
    the identifier for each cell.

    .. versionadded:: (cfdm) 1.11.0.0

    .. seealso:: `CellConnectivitySubarray`

    """

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        If a child class requires different subarray classes than the
        ones defined here, then they must be defined in the __new__
        method of the child class.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        instance = super().__new__(cls)
        instance._Subarray = {"cell connectivity": CellConnectivitySubarray}
        return instance

    def __init__(
        self,
        cell_connectivity=None,
        start_index=None,
        cell_dimension=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            cell_connectivity: array_like
                A 2-d integer array that contains indices which map
                each cell to its neighbours, as found in a UGRID
                "face_face_connectivity", or
                "volume_volume_connectivity" variable.

            {{init start_index: `int`}}

            {{init cell_dimension: `int`}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        if cell_connectivity is None:
            shape = None
        else:
            shape = cell_connectivity.shape
            if cell_dimension == 1:
                shape = shape[::-1]

            shape = (shape[0], shape[1] + 1)

        # Note: Setting compressed_dimensions={1: (1,)} means that
        #       only one subarray will ever span the trailing
        #       dimension of the 'cell_connectivity' array, but the
        #       leading dimension may be chunked.
        super().__init__(
            connectivity=cell_connectivity,
            shape=shape,
            compressed_dimensions={1: (1,)},
            start_index=start_index,
            cell_dimension=cell_dimension,
            compression_type="cell connectivity",
            source=source,
            copy=copy,
        )



from functools import reduce
from operator import mul

from ...abstract import Container
from ...utils import cached_property


class Array(Container):
    """Abstract base class for a container of an array.

    The form of the array is defined by the initialisation parameters
    of a subclass.

    See `cfdm.core.NumpyArray` for an example implementation.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

    def __array__(self, dtype=None, copy=None):
        """The numpy array interface.

        .. versionadded:: (cfdm) 1.8.7.0

        :Parameters:

            dtype: optional
                Typecode or data-type to which the array is cast.

            copy: `None` or `bool`
                Included to match the v2 `numpy.ndarray.__array__`
                API, but ignored. The return numpy array is always
                independent.

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        """
        array = self.array
        if dtype is None:
            return array

        return array.astype(dtype, copy=False)

    def __deepcopy__(self, memo):
        """Called by the `copy.deepcopy` function.

        x.__deepcopy__() <==> copy.deepcopy(x)

        Copy-on-write is employed. Therefore, after copying, care must be
        taken when making in-place modifications to attributes of either
        the original or the new copy.

        .. versionadded:: (cfdm) 1.7.0

        **Examples**

        >>> import copy
        >>> y = copy.deepcopy(x)

        """
        return self.copy()

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples**

        >>> n = a.array
        >>> isinstance(n, numpy.ndarray)
        True

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.array"
        )  # pragma: no cover

    @property
    def dtype(self):
        """Data-type of the array.

        .. versionadded:: (cfdm) 1.7.0

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.dtype"
        )  # pragma: no cover

    @cached_property
    def ndim(self):
        """Number of array dimensions.

        .. versionadded:: (cfdm) 1.7.0

        """
        return len(self.shape)

    @property
    def shape(self):
        """Shape of the array.

        .. versionadded:: (cfdm) 1.7.0

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.shape"
        )  # pragma: no cover

    @property
    def size(self):
        """Number of elements in the array.

        .. versionadded:: (cfdm) 1.7.0

        """
        return reduce(mul, self.shape, 1)

    def copy(self):
        """Return a deep copy of the array.

        ``a.copy() is equivalent to ``copy.deepcopy(a)``.

        Copy-on-write is employed. Therefore, after copying, care must be
        taken when making in-place modifications to attributes of either
        the original or the new copy.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> b = a.copy()

        """
        return type(self)(source=self, copy=True)
